@@ -2,7 +2,7 @@
 #
 # Workbook - A class for writing the Excel XLSX Workbook file.
 #
-# Copyright 2013-2015, John McNamara, jmcnamara@cpan.org
+# Copyright 2013-2016, John McNamara, jmcnamara@cpan.org
 #
 
 # Standard packages.
@@ -15,7 +15,7 @@ from datetime import datetime
 from zipfile import ZipFile, ZIP_DEFLATED
 from struct import unpack
 
-from .compatibility import str_types
+from .compatibility import int_types, num_types, str_types, force_unicode
 
 # Package imports.
 from . import xmlwriter
@@ -64,6 +64,7 @@ class Workbook(xmlwriter.XMLwriter):
         self.strings_to_numbers = options.get('strings_to_numbers', False)
         self.strings_to_formulas = options.get('strings_to_formulas', True)
         self.strings_to_urls = options.get('strings_to_urls', True)
+        self.nan_inf_to_errors = options.get('nan_inf_to_errors', False)
         self.default_date_format = options.get('default_date_format', None)
         self.optimization = options.get('constant_memory', False)
         self.in_memory = options.get('in_memory', False)
@@ -83,7 +84,7 @@ class Workbook(xmlwriter.XMLwriter):
         self.worksheets_objs = []
         self.charts = []
         self.drawings = []
-        self.sheetnames = []
+        self.sheetnames = {}
         self.formats = []
         self.xf_formats = []
         self.xf_format_indices = {}
@@ -96,7 +97,8 @@ class Workbook(xmlwriter.XMLwriter):
         self.named_ranges = []
         self.custom_colors = []
         self.doc_properties = {}
-        self.localtime = datetime.now()
+        self.custom_properties = []
+        self.createtime = datetime.utcnow()
         self.num_vml_files = 0
         self.num_comment_files = 0
         self.x_window = 240
@@ -146,6 +148,14 @@ class Workbook(xmlwriter.XMLwriter):
             raise Exception("Exception caught in workbook destructor. "
                             "Explicit close() may be required for workbook.")
 
+    def __enter__(self):
+        """Return self object to use with "with" statement."""
+        return self
+
+    def __exit__(self, type, value, traceback):
+        """Close workbook when exiting "with" statement."""
+        self.close()
+
     def add_worksheet(self, name=None):
         """
         Add a new worksheet to the Excel workbook.
@@ -264,7 +274,8 @@ class Workbook(xmlwriter.XMLwriter):
 
         """
         if not is_stream and not os.path.exists(vba_project):
-            warn("VBA project binary file '%s' not found." % vba_project)
+            warn("VBA project binary file '%s' not found."
+                 % force_unicode(vba_project))
             return -1
 
         self.vba_project = vba_project
@@ -272,7 +283,7 @@ class Workbook(xmlwriter.XMLwriter):
 
     def close(self):
         """
-        Call finalisation code and close file.
+        Call finalization code and close file.
 
         Args:
             None.
@@ -298,9 +309,55 @@ class Workbook(xmlwriter.XMLwriter):
         """
         self.doc_properties = properties
 
+    def set_custom_property(self, name, value, property_type=None):
+        """
+        Set a custom document property.
+
+        Args:
+            name:          The name of the custom property.
+            value:         The value of the custom property.
+            property_type: The type of the custom property. Optional.
+
+        Returns:
+            Nothing.
+
+        """
+        if name is None or value is None:
+            warn("The name and value parameters must be non-None in "
+                 "set_custom_property()")
+            return -1
+
+        if property_type is None:
+            # Determine the property type from the Python type.
+            if isinstance(value, bool):
+                property_type = 'bool'
+            elif isinstance(value, datetime):
+                property_type = 'date'
+            elif isinstance(value, int_types):
+                property_type = 'number_int'
+            elif isinstance(value, num_types):
+                property_type = 'number'
+            else:
+                property_type = 'text'
+
+        if property_type == 'date':
+            value = value.strftime("%Y-%m-%dT%H:%M:%SZ")
+
+        if property_type == 'text' and len(value) > 255:
+            warn("Length of 'value' parameter exceeds Excel's limit of 255 "
+                 "characters in set_custom_property(): '%s'" %
+                 force_unicode(value))
+
+        if len(name) > 255:
+            warn("Length of 'name' parameter exceeds Excel's limit of 255 "
+                 "characters in set_custom_property(): '%s'" %
+                 force_unicode(name))
+
+        self.custom_properties.append((name, value, property_type))
+
     def set_calc_mode(self, mode, calc_id=None):
         """
-        Set the Excel caclcuation mode for the workbook.
+        Set the Excel calculation mode for the workbook.
 
         Args:
             mode: String containing one of:
@@ -355,28 +412,31 @@ class Workbook(xmlwriter.XMLwriter):
 
             # Warn if the sheet index wasn't found.
             if sheet_index is None:
-                warn("Unknown sheet name '%s' in defined_name()" % sheetname)
+                warn("Unknown sheet name '%s' in defined_name()"
+                     % force_unicode(sheetname))
                 return -1
         else:
             # Use -1 to indicate global names.
             sheet_index = -1
 
         # Warn if the defined name contains invalid chars as defined by Excel.
-        if (not re.match(r'^[\w\\][\w.]*$', name, re.UNICODE)
+        if (not re.match(r'^[\w\\][\w\\.]*$', name, re.UNICODE)
                 or re.match(r'^\d', name)):
-            warn("Invalid Excel characters in defined_name(): '%s'" % name)
+            warn("Invalid Excel characters in defined_name(): '%s'"
+                 % force_unicode(name))
             return -1
 
         # Warn if the defined name looks like a cell name.
         if re.match(r'^[a-zA-Z][a-zA-Z]?[a-dA-D]?[0-9]+$', name):
-            warn("Name looks like a cell name in defined_name(): '%s'" % name)
+            warn("Name looks like a cell name in defined_name(): '%s'"
+                 % force_unicode(name))
             return -1
 
         # Warn if the name looks like a R1C1 cell reference.
         if (re.match(r'^[rcRC]$', name)
                 or re.match(r'^[rcRC]\d+[rcRC]\d+$', name)):
             warn("Invalid name '%s' like a RC cell ref in defined_name()"
-                 % name)
+                 % force_unicode(name))
             return -1
 
         self.defined_names.append([name, sheet_index, formula, False])
@@ -394,6 +454,19 @@ class Workbook(xmlwriter.XMLwriter):
         """
         return self.worksheets_objs
 
+    def get_worksheet_by_name(self, name):
+        """
+        Return a worksheet object in the workbook using the sheetname.
+
+        Args:
+            name: The name of the worksheet.
+
+        Returns:
+            A worksheet object or None.
+
+        """
+        return self.sheetnames.get(name)
+
     def use_zip64(self):
         """
         Allow ZIP64 extensions when writing xlsx file zip container.
@@ -536,7 +609,7 @@ class Workbook(xmlwriter.XMLwriter):
         sheet_index = len(self.worksheets_objs)
         name = self._check_sheetname(name, is_chartsheet)
 
-        # Initialisation data to pass to the worksheet.
+        # Initialization data to pass to the worksheet.
         init_data = {
             'name': name,
             'index': sheet_index,
@@ -548,6 +621,7 @@ class Workbook(xmlwriter.XMLwriter):
             'strings_to_numbers': self.strings_to_numbers,
             'strings_to_formulas': self.strings_to_formulas,
             'strings_to_urls': self.strings_to_urls,
+            'nan_inf_to_errors': self.nan_inf_to_errors,
             'default_date_format': self.default_date_format,
             'default_url_format': self.default_url_format,
             'excel2003_style': self.excel2003_style,
@@ -561,7 +635,7 @@ class Workbook(xmlwriter.XMLwriter):
         worksheet._initialize(init_data)
 
         self.worksheets_objs.append(worksheet)
-        self.sheetnames.append(name)
+        self.sheetnames[name] = worksheet
 
         return worksheet
 
@@ -767,7 +841,7 @@ class Workbook(xmlwriter.XMLwriter):
         fills['0:0:0'] = 0
         fills['17:0:0'] = 1
 
-        # Store the DXF colours separately since them may be reversed below.
+        # Store the DXF colors separately since them may be reversed below.
         for xf_format in self.dxf_formats:
             if xf_format.pattern or xf_format.bg_color or xf_format.fg_color:
                 xf_format.has_dxf_fill = 1
@@ -776,10 +850,10 @@ class Workbook(xmlwriter.XMLwriter):
 
         for xf_format in self.xf_formats:
             # The following logical statements jointly take care of special
-            # cases in relation to cell colours and patterns:
+            # cases in relation to cell colors and patterns:
             # 1. For a solid fill (_pattern == 1) Excel reverses the role of
-            # foreground and background colours, and
-            # 2. If the user specifies a foreground or background colour
+            # foreground and background colors, and
+            # 2. If the user specifies a foreground or background color
             # without a pattern they probably wanted a solid fill, so we fill
             # in the defaults.
             if (xf_format.pattern == 1 and xf_format.bg_color != 0
@@ -859,21 +933,21 @@ class Workbook(xmlwriter.XMLwriter):
         # Sort the list of list of internal and user defined names in
         # the same order as used by Excel.
 
-        # Add a normalise name string to each list for sorting.
+        # Add a normalize name string to each list for sorting.
         for name_list in names:
             (defined_name, _, sheet_name, _) = name_list
 
-            # Normalise the defined name by removing any leading '_xmln.'
+            # Normalize the defined name by removing any leading '_xmln.'
             # from internal names and lowercasing the string.
             defined_name = defined_name.replace('_xlnm.', '').lower()
 
-            # Normalise the sheetname by removing the leading quote and
+            # Normalize the sheetname by removing the leading quote and
             # lowercasing the string.
             sheet_name = sheet_name.lstrip("'").lower()
 
             name_list.append(defined_name + "::" + sheet_name)
 
-        # Sort based on the normalised key.
+        # Sort based on the normalized key.
         names.sort(key=operator.itemgetter(4))
 
         # Remove the extra key used for sorting.
@@ -924,6 +998,10 @@ class Workbook(xmlwriter.XMLwriter):
                 sheet._prepare_image(index, image_ref_id, drawing_id, width,
                                      height, name, image_type, x_dpi, y_dpi)
 
+            # Prepare the worksheet shapes.
+            for index in range(shape_count):
+                sheet._prepare_shape(index, drawing_id)
+
             # Prepare the header images.
             for index in range(header_image_count):
 
@@ -1029,6 +1107,12 @@ class Workbook(xmlwriter.XMLwriter):
         if not image_data:
             fh.close()
 
+        # Set a default dpi for images with 0 dpi.
+        if x_dpi == 0:
+            x_dpi = 96
+        if y_dpi == 0:
+            y_dpi = 96
+
         return image_type, width, height, image_name, x_dpi, y_dpi
 
     def _process_png(self, data):
@@ -1112,6 +1196,12 @@ class Workbook(xmlwriter.XMLwriter):
                     x_dpi = x_density * 2.54
                     y_dpi = y_density * 2.54
 
+                # Workaround for incorrect dpi.
+                if x_dpi == 1:
+                    x_dpi = 96
+                if y_dpi == 1:
+                    y_dpi = 96
+
             if marker == 0xFFDA:
                 end_marker = True
                 continue
@@ -1161,7 +1251,7 @@ class Workbook(xmlwriter.XMLwriter):
         sheetname = sheetname.strip("'")
 
         if sheetname in self.sheetnames:
-            return self.sheetnames.index(sheetname)
+            return self.sheetnames[sheetname].index
         else:
             return None
 
@@ -1229,6 +1319,7 @@ class Workbook(xmlwriter.XMLwriter):
     def _prepare_tables(self):
         # Set the table ids for the worksheet tables.
         table_id = 0
+        seen = {}
 
         for sheet in self.worksheets():
             table_count = len(sheet.tables)
@@ -1236,20 +1327,27 @@ class Workbook(xmlwriter.XMLwriter):
             if not table_count:
                 continue
 
-            sheet._prepare_tables(table_id + 1)
+            sheet._prepare_tables(table_id + 1, seen)
             table_id += table_count
 
     def _add_chart_data(self):
-        # Add "cached" data to charts to provide the numCache and strCacher
+        # Add "cached" data to charts to provide the numCache and strCache
         # data for series and title/axis ranges.
         worksheets = {}
         seen_ranges = {}
+        charts = []
 
         # Map worksheet names to worksheet objects.
         for worksheet in self.worksheets():
             worksheets[worksheet.name] = worksheet
 
+        # Build a list of the worksheet charts including any combined charts.
         for chart in self.charts:
+            charts.append(chart)
+            if chart.combined:
+                charts.append(chart.combined)
+
+        for chart in charts:
 
             for c_range in chart.formula_ids.keys():
                 r_id = chart.formula_ids[c_range]
@@ -1286,7 +1384,8 @@ class Workbook(xmlwriter.XMLwriter):
                 # in a chart series formula.
                 if sheetname not in worksheets:
                     warn("Unknown worksheet reference '%s' in range "
-                         "'%s' passed to add_series()" % (sheetname, c_range))
+                         "'%s' passed to add_series()"
+                         % (force_unicode(sheetname), force_unicode(c_range)))
                     chart.formula_data[r_id] = []
                     seen_ranges[c_range] = []
                     continue
@@ -1315,7 +1414,7 @@ class Workbook(xmlwriter.XMLwriter):
             sheetname = c_range[:pos]
             cells = c_range[pos + 1:]
         else:
-            return None
+            return None, None
 
         # Split the cell range into 2 cells or else use single cell for both.
         if cells.find(':') > 0:
@@ -1327,12 +1426,17 @@ class Workbook(xmlwriter.XMLwriter):
         sheetname = sheetname.strip("'")
         sheetname = sheetname.replace("''", "'")
 
-        (row_start, col_start) = xl_cell_to_rowcol(cell_1)
-        (row_end, col_end) = xl_cell_to_rowcol(cell_2)
+        try:
+            # Get the row, col values from the Excel ranges. We do this in a
+            # try block for ranges that can't be parsed such as defined names.
+            (row_start, col_start) = xl_cell_to_rowcol(cell_1)
+            (row_end, col_end) = xl_cell_to_rowcol(cell_2)
+        except:
+            return None, None
 
-        # Check that we have a 1D range only.
+        # We only handle 1D ranges.
         if row_start != row_end and col_start != col_end:
-            return None
+            return None, None
 
         return sheetname, [row_start, col_start, row_end, col_end]
 
@@ -1498,7 +1602,7 @@ class Workbook(xmlwriter.XMLwriter):
 class WorksheetMeta(object):
     """
     A class to track worksheets data such as the active sheet and the
-    first sheet..
+    first sheet.
 
     """
 
